from dataclasses import dataclass, field
from typing import List, Mapping, Union, Optional

from annotell.input_api.model.ego import EgoVehiclePose
from annotell.input_api.model.input.abstract.sequence_frame import SequenceFrame
from annotell.input_api.model.input.resources import PointCloud, Image, VideoFrame


@dataclass
class Frame(SequenceFrame):
    point_clouds: List[PointCloud] = field(default_factory=list)
    ego_vehicle_pose: Optional[EgoVehiclePose] = None
    metadata: Mapping[str, Union[int, float, str, bool]] = field(default_factory=dict)

    def to_dict(self) -> dict:
        return dict(
            frameId=self.frame_id,
            relativeTimestamp=self.relative_timestamp,
            pointClouds=[pc.to_dict() for pc in self.point_clouds] if self.point_clouds else None,
            egoVehiclePose=self.ego_vehicle_pose.to_dict() if self.ego_vehicle_pose else None,
            metadata=self.metadata
        )



from dataclasses import field, dataclass
from typing import Union, Mapping

from annotell.input_api.model.input.lidars_and_cameras.frame import Frame


@dataclass
class Lidars:
    external_id: str
    frame: Frame
    metadata: Mapping[str, Union[int, float, str, bool]] = field(default_factory=dict)

    def to_dict(self) -> dict:
        return dict(frame=self.frame.to_dict(),
                    externalId=self.external_id,
                    metadata=self.metadata)



from dataclasses import dataclass, field

from typing import List, Mapping, Union
from annotell.input_api.model.input.lidars_sequence.frame import Frame


@dataclass
class LidarsSequence:
    external_id: str
    frames: List[Frame]
    metadata: Mapping[str, Union[int, float, str]] = field(default_factory=dict)

    def to_dict(self) -> dict:
        return dict(frames=[frame.to_dict() for frame in self.frames],
                    externalId=self.external_id,
                    metadata=self.metadata)



from dataclasses import dataclass
from typing import Dict, Optional
from enum import Enum

from deprecated import deprecated

from annotell.input_api.model.abstract.abstract_models import RequestCall


class CameraType(str, Enum):
    PINHOLE = "pinhole"
    FISHEYE = "fisheye"
    KANNALA = "kannala"


@dataclass
class RotationQuaternion(RequestCall):
    w: float
    x: float
    y: float
    z: float

    def to_dict(self) -> Dict:
        return self.__dict__

    @staticmethod
    def from_json(js: dict):
        return RotationQuaternion(
            w=js["w"],
            x=js["x"],
            y=js["y"],
            z=js["z"],
        )


@dataclass
class Position(RequestCall):
    x: float
    y: float
    z: float

    def to_dict(self) -> Dict:
        return self.__dict__

    @staticmethod
    def from_json(js: dict):
        return Position(
            x=js["x"],
            y=js["y"],
            z=js["z"],
        )


@dataclass
class CameraMatrix(RequestCall):
    fx: float
    fy: float
    cx: float
    cy: float

    def to_dict(self) -> Dict:
        return self.__dict__

    @staticmethod
    def from_json(js: dict):
        return CameraMatrix(
            fx=js["fx"],
            fy=js["fy"],
            cx=js["cx"],
            cy=js["cy"],
        )


@dataclass
class DistortionCoefficients(RequestCall):
    k1: float
    k2: float
    p1: float
    p2: float
    k3: Optional[float] = None

    def to_dict(self) -> Dict:
        return {k: v for k, v in self.__dict__.items() if v is not None}

    def validate(self, camera_type: CameraType):
        if camera_type == CameraType.KANNALA:
            assert self.k3 is None
        else:
            assert self.k3 is not None

    @staticmethod
    def from_json(js: dict):
        return DistortionCoefficients(k1=js["k1"], k2=js["k2"], p1=js["p1"], p2=js["p2"], k3=js.get("k3"))


@dataclass
class UndistortionCoefficients(RequestCall):
    l1: float
    l2: float
    l3: float
    l4: float

    def to_dict(self) -> Dict:
        return self.__dict__

    @staticmethod
    def from_json(js: dict):
        if js:
            return UndistortionCoefficients(l1=js["l1"], l2=js["l2"], l3=js["l3"], l4=js["l4"])
        else:
            return None


@deprecated(reason="Will be removed, migrate to typed calibrations", action="once")
@dataclass
class CameraProperty(RequestCall):
    camera_type: CameraType

    def to_dict(self):
        return {"camera_type": self.camera_type}

    def get_camera_type(self) -> CameraType:
        return self.camera_type

    @staticmethod
    def from_json(js: dict):
        return CameraProperty(camera_type=js["camera_type"])


@deprecated(reason="Will be removed, migrate to typed calibrations", action="once")
@dataclass
class LidarCalibration(RequestCall):
    position: Position
    rotation_quaternion: RotationQuaternion

    def to_dict(self) -> Dict:
        return {"position": self.position.to_dict(), "rotation_quaternion": self.rotation_quaternion.to_dict()}

    @staticmethod
    def from_json(js: dict):
        return LidarCalibration(
            position=Position.from_json(js["position"]), rotation_quaternion=RotationQuaternion.from_json(js["rotation_quaternion"])
        )


@deprecated(reason="Will be removed, migrate to typed calibrations", action="once")
@dataclass
class CameraCalibration(RequestCall):
    position: Position
    rotation_quaternion: RotationQuaternion
    camera_matrix: CameraMatrix
    distortion_coefficients: DistortionCoefficients
    camera_properties: CameraProperty
    image_height: int
    image_width: int
    undistortion_coefficients: Optional[UndistortionCoefficients] = None

    def __post_init__(self):
        camera_type = self.camera_properties.get_camera_type()
        self.distortion_coefficients.validate(camera_type=camera_type)
        if camera_type == CameraType.KANNALA:
            assert self.undistortion_coefficients is not None

    def to_dict(self) -> Dict:
        base = {
            "position": self.position.to_dict(),
            "rotation_quaternion": self.rotation_quaternion.to_dict(),
            "camera_matrix": self.camera_matrix.to_dict(),
            "camera_properties": self.camera_properties.to_dict(),
            "distortion_coefficients": self.distortion_coefficients.to_dict(),
            "image_height": self.image_height,
            "image_width": self.image_width
        }
        if self.undistortion_coefficients is not None:
            base["undistortion_coefficients"] = self.undistortion_coefficients.to_dict()

        return base

    @staticmethod
    def from_json(js: dict):
        return CameraCalibration(
            position=Position.from_json(js["position"]),
            rotation_quaternion=RotationQuaternion.from_json(js["rotation_quaternion"]),
            camera_matrix=CameraMatrix.from_json(js["camera_matrix"]),
            distortion_coefficients=DistortionCoefficients.from_json(js["distortion_coefficients"]),
            camera_properties=CameraProperty.from_json(js["camera_properties"]),
            image_height=js["image_height"],
            image_width=js["image_width"],
            undistortion_coefficients=UndistortionCoefficients.from_json(js.get("undistortion_coefficients")),
        )



"""Input API model"""

from .abstract import *
from .annotation import *
from .calibration import *
from .files_to_upload import *
from .input import *
from .projects import *
from .upload_url import *
from .ego import *



import logging
from typing import Optional, List

import annotell.input_api.model.input as InputModel
from annotell.input_api.resources.abstract import CreateableInputAPIResource

log = logging.getLogger(__name__)


class LidarsAndCamerasSequence(CreateableInputAPIResource):

    path = 'lidars-and-cameras-sequence'

    def create(self,
               lidars_and_cameras_sequence: InputModel.LidarsAndCamerasSequence,
               project: Optional[str] = None,
               batch: Optional[str] = None,
               annotation_types: Optional[List[str]] = None,
               dryrun: bool = False) -> Optional[InputModel.CreateInputResponse]:
        """
        Upload files and create an input of type ``LidarsAndCamerasSequence``.

        :param lidars_and_cameras_sequence: class containing 2D and 3D resources that constitute the input
        :param project: project to add input to
        :param batch: batch, defaults to latest open batch
        :param annotation_types: annotation types for which to produce annotations for. Defaults to `None` (corresponds to all available annotation types). Passing an empty list will result in the same behaviour as passing `None`.
        :param dryrun: If True the files/metadata will be validated but no input job will be created.
        :returns InputJobCreated: Class containing id of the created input job, or `None` if dryrun.
        """

        payload = lidars_and_cameras_sequence.to_dict()

        response = self._post_input_request(self.path, payload,
                                            project=project,
                                            batch=batch,
                                            annotation_types=annotation_types,
                                            dryrun=dryrun)

        if dryrun:
            return None

        log.info(f"Created inputs for files with uuid={response.input_uuid}")
        return InputModel.CreateInputResponse.from_input_job_response(response)



from datetime import datetime
from typing import Dict, Mapping, Optional, Union
from dataclasses import dataclass

from annotell.input_api.model.abstract.abstract_models import Response
from annotell.input_api.util import ts_to_dt
from annotell.input_api.model.calibration import sensors as CM
from annotell.input_api.model.calibration.common import BaseCalibration, CalibrationType
from annotell.input_api.model.calibration.lidar.lidar_calibration import LidarCalibration
from annotell.input_api.model.calibration.camera.kannala_calibration import KannalaCalibration
from annotell.input_api.model.calibration.camera.fisheye_calibration import FisheyeCalibration
from annotell.input_api.model.calibration.camera.pinhole_calibration import PinholeCalibration


@dataclass
class SensorCalibration:
    external_id: str
    calibration: Dict[str, Union[BaseCalibration, CM.CameraCalibration]]

    def to_dict(self):
        return {'externalId': self.external_id, 'calibration': {k: v.to_dict() for (k, v) in self.calibration.items()}}


@dataclass
class SensorCalibrationEntry(Response):
    id: str
    external_id: str
    created: datetime
    calibration: Optional[Mapping[str, Union[BaseCalibration, CM.CameraCalibration]]]

    @classmethod
    def from_json(cls, js: dict):
        calibrations = js.get("calibration", {})

        calibration = {}
        for sensor, calib in calibrations.items():
            calibration[sensor] = cls._parse_calibration(calib)
        return SensorCalibrationEntry(id=js["id"], external_id=js["externalId"], created=ts_to_dt(js["created"]), calibration=calibration)

    @staticmethod
    def _parse_calibration(calibration: Dict) -> Union[BaseCalibration, CM.CameraCalibration]:

        calibration_type = calibration.get("calibration_type")
        if calibration_type is None:
            if calibration.get("camera_matrix") is None:
                return LidarCalibration.parse_obj(calibration)
            return CM.CameraCalibration.from_json(calibration)

        if calibration_type == CalibrationType.PINHOLE:
            return PinholeCalibration.parse_obj(calibration)

        if calibration_type == CalibrationType.KANNALA:
            return KannalaCalibration.parse_obj(calibration)

        if calibration_type == CalibrationType.FISHEYE:
            return FisheyeCalibration.parse_obj(calibration)

        if calibration_type == CalibrationType.LIDAR:
            return LidarCalibration.parse_obj(calibration)

        raise TypeError(f"Unable to parse calibration type: {calibration_type}")



from typing import Optional, Dict
from datetime import datetime
from dataclasses import dataclass
from annotell.input_api.model.abstract.abstract_models import Response
from annotell.input_api.util import ts_to_dt

@dataclass
class Annotation(Response):
    input_uuid: str
    annotation_type: str
    created: datetime
    content: Dict

    @staticmethod
    def from_json(js: dict):
        return Annotation(
                input_uuid=js["inputUuid"],
                annotation_type=js["annotationType"],
                created=ts_to_dt(js["created"]),
                content=js["content"])

@dataclass
class PartialAnnotation(Response):
    input_uuid: str
    annotation_type: str
    created: datetime
    uri: str

    @staticmethod
    def from_json(js: dict):
        return PartialAnnotation(
            input_uuid=js["inputUuid"],
            annotation_type=js["annotationType"],
            created=ts_to_dt(js["created"]),
            uri=js["uri"])

    def to_annotation(self, content: Dict) -> Annotation:
        return Annotation(
            input_uuid=self.input_uuid,
            annotation_type=self.annotation_type,
            created=self.created,
            content=content)



from typing import List, Dict, Optional, Generator
from deprecated import deprecated
from annotell.input_api.model.annotation import ExportAnnotation
from annotell.input_api.model.annotation.client_annotation import Annotation, PartialAnnotation
from annotell.input_api.util import filter_none
from annotell.input_api.resources.abstract import InputAPIResource


class AnnotationResource(InputAPIResource):
    @deprecated(reason="Returns annotations in client-specific format."
                       "This method is deprecated in favour of `get_project_annotations` which lets you list "
                       "annotations based on project, batch and annotation type.")
    def get_annotations(
            self,
            input_uuids: List[str]
    ) -> Dict[str, List[ExportAnnotation]]:
        """
        Returns the export ready annotations, either

        :param input_uuids: List with input uuid
        :return Dict: A dictionary containing the ready annotations
        """
        external_id_query_param = ",".join(input_uuids) if input_uuids else None
        json_resp = self._client.get("v1/annotations", params=filter_none({
            "inputs": external_id_query_param
        }))

        annotations = dict()
        for k, v in json_resp.items():
            annotations[k] = [
                ExportAnnotation.from_json(annotation) for annotation in v
            ]
        return annotations

    def get_project_annotations(self,
                                project: str,
                                annotation_type: str,
                                batch: Optional[str] = None) -> Generator[Annotation, None, None]:
        url = f"v1/annotations/projects/{project}/"
        if batch:
            url += f"batch/{batch}/"

        url += f"annotation-type/{annotation_type}"

        annotations = self._client.get(url)
        for js in annotations:
            partial_annotation = PartialAnnotation.from_json(js)
            content = self.get_json(partial_annotation.uri)
            yield partial_annotation.to_annotation(content)

    def get_annotation(self,
                       input_uuid: str,
                       annotation_type: str) -> Annotation:
        json_resp = self._client.get(f"v1/annotations/inputs/{input_uuid}/annotation-type/{annotation_type}")
        annotation = Annotation.from_json(json_resp)
        return annotation



from annotell.input_api.model.input.resources.resource import Resource
from typing import Optional
from abc import ABC
from dataclasses import dataclass
from annotell.input_api.util import filter_none

camera_sensor_default = "CAM"

@dataclass
class VideoTS(ABC):
    video_timestamp: int

@dataclass
class VideoFrame(Resource, VideoTS):
    filename: str
    resource_id: Optional[str] = None
    sensor_name: str = camera_sensor_default
    video_timestamp: int

    def to_dict(self) -> dict:
        return filter_none({
            "filename": self.filename,
            "videoTimestamp": self.video_timestamp,
            "sensorName": self.sensor_name,
            "resourceId": self.resolve_resource_id()
        })



from enum import Enum
from typing import Dict

from pydantic import BaseModel


class CalibrationType(str, Enum):
    PINHOLE = "pinhole"
    FISHEYE = "fisheye"
    KANNALA = "kannala"
    LIDAR = "lidar"


class BaseSerializer(BaseModel):

    @classmethod
    def from_json(cls, js: Dict):
        return cls.parse_obj(js)

    def to_dict(self) -> Dict:
        return self.dict(exclude_none=True)


class RotationQuaternion(BaseSerializer):
    w: float
    x: float
    y: float
    z: float


class Position(BaseSerializer):
    x: float
    y: float
    z: float


class BaseCalibration(BaseSerializer):
    position: Position
    rotation_quaternion: RotationQuaternion



from typing import Dict
from dataclasses import dataclass
from annotell.input_api.model.abstract.abstract_models import Response


@dataclass
class UploadUrls(Response):
    files_to_url: Dict[str, str]
    input_uuid: int

    @staticmethod
    def from_json(js: dict):
        return UploadUrls(
            # property names does not match.
            files_to_url=js["files"],
            input_uuid=js["jobId"]
        )



from pathlib import Path
from abc import ABC
from typing import Optional
from dataclasses import dataclass


@dataclass
class Resource(ABC):
    filename: str
    resource_id: Optional[str]
    sensor_name: str

    def resolve_resource_id(self):
        return self.resource_id or str(Path(self.filename).expanduser())



from dataclasses import dataclass
from typing import Optional, List
from annotell.input_api.model.abstract.abstract_models import Response
from datetime import datetime
from enum import Enum
from annotell.input_api.util import ts_to_dt


class InputStatus(str, Enum):
    Processing = "processing"
    Created = "created"
    Failed = "failed"
    InvalidatedBadContent = "invalidated:broken-input"
    InvalidatedSlamRerun = "invalidated:slam-rerun"
    InvalidatedDuplicate = "invalidated:duplicate"
    InvalidatedIncorrectlyCreated = "invalidated:incorrectly-created"


@dataclass
class Input(Response):
    uuid: str
    external_id: str
    batch: str
    input_type: str
    status: InputStatus
    created: datetime
    calibration_id: Optional[str]
    view_link: Optional[str]
    error_message: Optional[str]

    @staticmethod
    def from_json(js: dict):
        return Input(
            uuid=js["internalId"],
            external_id=js["externalId"],
            batch=js["batchId"],
            input_type=js["inputType"],
            status=js["status"],
            created=ts_to_dt(js["created"]),
            view_link=js.get("viewLink"),
            calibration_id=js.get("calibrationId"),
            error_message=js.get("errorMessage")
        )



from annotell.input_api.file_resource_client import FileResourceClient
from annotell.input_api.http_client import HttpClient


class InputAPIResource(FileResourceClient):

    def __init__(self, client: HttpClient):
        super().__init__()
        self._client = client



from dataclasses import dataclass
from typing import Optional

from annotell.input_api.util import filter_none
from annotell.input_api.model.input.resources.resource import Resource

lidar_sensor_default = "lidar"


@dataclass
class PointCloud(Resource):
    filename: str
    resource_id: Optional[str] = None
    sensor_name: str = lidar_sensor_default

    def to_dict(self) -> dict:
        return filter_none({
            "filename": self.filename,
            "resourceId": self.resolve_resource_id(),
            "sensorName": self.sensor_name
        })



from dataclasses import dataclass
from datetime import datetime

from annotell.input_api.model.abstract.abstract_models import Response
from annotell.input_api.util import ts_to_dt


@dataclass
class Project(Response):
    created: datetime
    title: str
    description: str
    status: str
    project: str

    @ staticmethod
    def from_json(js: dict):
        return Project(
            created=ts_to_dt(js["created"]),
            title=js["title"],
            description=js["description"],
            status=js["status"],
            project=js["project"]
        )



from dataclasses import dataclass
from typing import List


@dataclass
class FilesToUpload:
    """
    Used when retrieving upload urls from input api
    """

    files_to_upload: List[str]

    def to_dict(self):
        return dict(filesToUpload=self.files_to_upload)



from dataclasses import dataclass
from typing import List, Optional

from dataclasses import dataclass, field

from typing import List, Mapping, Union
from annotell.input_api.model.input.lidars_and_cameras_sequence.frame import Frame
from annotell.input_api.model.input.sensor_specification import SensorSpecification


@dataclass
class LidarsAndCamerasSequence:
    external_id: str
    frames: List[Frame]
    calibration_id: str
    sensor_specification: Optional[SensorSpecification] = None
    metadata: Mapping[str, Union[int, float, str]] = field(default_factory=dict)

    def to_dict(self) -> dict:
        return dict(frames=[frame.to_dict() for frame in self.frames],
                    sensorSpecification=self.sensor_specification.to_dict() if isinstance(self.sensor_specification, SensorSpecification) else None,
                    externalId=self.external_id,
                    calibrationId=self.calibration_id,
                    metadata=self.metadata)



from annotell.input_api.model.calibration.camera.common import BaseCameraCalibration, DistortionCoefficients
from annotell.input_api.model.calibration.common import CalibrationType


class FisheyeCalibration(BaseCameraCalibration):
    calibration_type = CalibrationType.FISHEYE.value
    distortion_coefficients: DistortionCoefficients
    xi: float
